//---------------------------------------------------------------------------
//  MCommon unit.
//  Copyright (c) 2003 Mental Studio - http://mental.mentsu.com
//  Author : Raptor - raptor@mentsu.com
//  Create : Sep.8-03
//  Last modify : Nov.9-03

//  TMauto_ptr :
//    derived from STL's auto_ptr template class, add a type cast operator
//  TMCDLL :
//    DLL managenment(auto load/unload, get procedure address etc.
//  TMCDLLProc :
//    DLL procedure management.
//---------------------------------------------------------------------------

#ifndef MCommonH
#define MCommonH

#include <windows.h>

#include <memory>

#ifdef __BORLANDC__

#include <vcl.h>

#else

#include <string>

typedef std::string AnsiString;

#endif

//---------------------------------------------------------------------------

template<class T>
class TMauto_ptr : public std::auto_ptr<T>
{
public:
  TMauto_ptr( T * src) : std::auto_ptr<T>( src ) { }

  operator T* ( ) const { return get( ); }
};

////////////////////////////////////////////////////
// TMCDLL: Wrapper class encapsulating a DLL
////////////////////////////////////////////////////

class TMCDLL
{
protected:
  AnsiString  FFileName;
  HINSTANCE   FHandle;

public:
  TMCDLL( AnsiString aFileName)
    : FFileName( aFileName ), FHandle( ::LoadLibrary( aFileName.c_str( ) ) ) { }
 ~TMCDLL( )
  {
    if ( FHandle )
      ::FreeLibrary( FHandle );
  }

  operator HINSTANCE( ) const
  {
    return FHandle;
  }
  FARPROC  GetProcAddress( AnsiString aFuncName ) const
  {
    return ::GetProcAddress( FHandle, aFuncName.c_str( ) );
  }

#ifdef __BORLANDC__
  __property AnsiString FileName = { read = FFileName };
  __property HINSTANCE  Handle   = { read = FHandle   };
#endif

private:
  TMCDLL( const TMCDLL& );
  TMCDLL& operator=( const TMCDLL& );
};

//////////////////////////////////////////////////
// TMCDLLProc: Encapsulates a DLL EntryPoint/FARPROC
//////////////////////////////////////////////////
template<class T> 
class TMCDLLProc
{
public:
  TMCDLLProc( TMCDLL * dll, AnsiString aFuncName )
  {
    FProc = dll->GetProcAddress( aFuncName );
  }
  operator bool( )    { return FProc != 0; }
  operator T( ) const { return ( T )FProc; }

protected:
  FARPROC   FProc;
};

//---------------------------------------------------------------------------
#endif
//---------------------------------------------------------------------------


 